<?php

declare(strict_types=1);

namespace Drupal\invoice_ninja;

use Drupal\Core\Config\ConfigFactoryInterface;
use Drupal\Core\Config\ImmutableConfig;
use Drupal\Core\KeyValueStore\KeyValueFactoryInterface;
use Drupal\invoice_ninja\Form\Settings;
use InvoiceNinja\Sdk\InvoiceNinja;

/**
 * Provides services to synchronize data with Invoice Ninja.
 */
final class InvoiceNinjaSynchronizer {

  protected const INVOICE_NINJA_COLLECTION_CLIENT = 'invoice_ninja.sync_client';

  protected const INVOICE_NINJA_COLLECTION_INVOICE = 'invoice_ninja.sync_invoice';

  protected const INVOICE_NINJA_COLLECTION_VAT = 'invoice_ninja.sync_vat';

  /**
   * The config.
   *
   * @var \Drupal\Core\Config\ImmutableConfig
   */
  protected ImmutableConfig $config;

  /**
   * The key/value factory service.
   *
   * @var \Drupal\Core\KeyValueStore\KeyValueFactoryInterface
   */
  protected KeyValueFactoryInterface $keyValueFactory;

  /**
   * The Invoice Ninja client.
   *
   * @var \InvoiceNinja\Sdk\InvoiceNinja|null
   */
  private ?InvoiceNinja $invoiceNinja;

  /**
   * The mapping of countries.
   *
   * @var array
   */
  private array $countryMapper = [];

  /**
   * The mapping of currencies.
   *
   * @var array
   */
  private array $currencyMapper = [];

  /**
   * Constructs an InvoiceNinjaApi object.
   *
   * @param \Drupal\Core\Config\ConfigFactoryInterface $configFactory
   *   The config factory.
   * @param \Drupal\Core\KeyValueStore\KeyValueFactoryInterface $keyValueFactory
   *   The key/value factory.
   */
  public function __construct(ConfigFactoryInterface $configFactory, KeyValueFactoryInterface $keyValueFactory) {
    $this->config = $configFactory->get(Settings::CONFIG_NAME);
    $this->keyValueFactory = $keyValueFactory;
  }

  /**
   * Return the status if the entity has already been synced.
   *
   * @param \Drupal\invoice_ninja\SyncableInterface $entity
   *   The entity.
   *
   * @return bool
   *   TRUE, if the entity has already been synced, FALSE otherwise.
   */
  public function isSynced(SyncableInterface $entity): bool {
    return match (TRUE) {
      $entity instanceof SyncableClientInterface => $this->keyValueFactory->get(self::INVOICE_NINJA_COLLECTION_CLIENT)
        ->get($entity->id()) !== NULL,
      $entity instanceof SyncableInvoiceInterface => $this->keyValueFactory->get(self::INVOICE_NINJA_COLLECTION_INVOICE)
        ->get($entity->id()) !== NULL,
      $entity instanceof SyncableVatInterface => $this->keyValueFactory->get(self::INVOICE_NINJA_COLLECTION_VAT)
        ->get($entity->id()) !== NULL,
      default => FALSE,
    };
  }

  /**
   * Syncs the client with Invoice Ninja.
   *
   * If the client already exists, an update gets performed.
   *
   * @param \Drupal\invoice_ninja\SyncableClientInterface $client
   *   The client.
   *
   * @return string
   *   The remote ID from Invoice Ninja.
   */
  public function syncClient(SyncableClientInterface $client): string {
    $clientData = $this->keyValueFactory->get(self::INVOICE_NINJA_COLLECTION_CLIENT)
      ->get($client->id());
    if (!$clientData) {
      $remoteClient = $this->getInvoiceNinjaClient()->clients->create(
        $this->getClientData($client)
      );
      $remoteId = $remoteClient['data']['id'];
      $this->storeId(self::INVOICE_NINJA_COLLECTION_CLIENT,
        $client->id(), $remoteId);
    }
    else {
      $lastSync = $clientData['last_sync'];
      $remoteId = $clientData['remote_id'];
      // @phpstan-ignore-next-line
      if ($client->getChangedTime() > $lastSync) {
        $this->getInvoiceNinjaClient()->clients->update($remoteId,
          $this->getClientData($client));
        $this->storeId(self::INVOICE_NINJA_COLLECTION_CLIENT,
          $client->id(), $remoteId);
      }
    }
    return $remoteId;
  }

  /**
   * Stores the given data, id and current timestamp into the given collection.
   *
   * @param string $collection
   *   The collection of the key/value store.
   * @param string $id
   *   The ID to store.
   * @param string $remoteId
   *   The remote ID of Invoice Ninja.
   */
  private function storeId(string $collection, string $id, string $remoteId): void {
    $data = [
      'remote_id' => $remoteId,
      'last_sync' => time(),
    ];
    $this->keyValueFactory->get($collection)->set($id, $data);
  }

  /**
   * Gets the data for the API call for a client.
   *
   * @param \Drupal\invoice_ninja\SyncableClientInterface $client
   *   The client.
   *
   * @return array
   *   The client data for the API.
   */
  private function getClientData(SyncableClientInterface $client): array {
    return [
      'name' => $client->getAddress()->getOrganization(),
      'address1' => $client->getAddress()->getAddressLine1(),
      'address2' => $client->getAddress()->getAddressLine2(),
      'postal_code' => $client->getAddress()->getPostalCode(),
      'city' => $client->getAddress()->getLocality(),
      'phone' => $client->getPhone(),
      'country_id' => $this->getCountryMapping()[$client->getAddress()->getCountryCode()],
      'settings' => [
        'currency_id' => $this->getCurrencyMapping()[$client->getCurrency()],
      ],
      'contacts' => $this->getContactsData($client),
    ];
  }

  /**
   * Gets the data for the API call for the contacts.
   *
   * @param \Drupal\invoice_ninja\SyncableClientInterface $client
   *   The client.
   *
   * @return array
   *   The contacts data for the API.
   */
  private function getContactsData(SyncableClientInterface $client): array {
    $contacts = [];
    /** @var \Drupal\invoice_ninja\SyncableContactInterface $contact */
    foreach ($client->getContacts() as $contact) {
      $contacts[] = [
        'first_name' => $contact->getGivenName(),
        'last_name' => $contact->getFamilyName(),
        'phone' => $contact->getPhone(),
        'email' => $contact->getEmail(),
      ];
    }
    return $contacts;
  }

  /**
   * Syncs the VAT with Invoice Ninja.
   *
   * @param \Drupal\invoice_ninja\SyncableVatInterface $vat
   *   The VAT.
   */
  public function syncVat(SyncableVatInterface $vat): void {
    $vatData = $this->keyValueFactory->get(self::INVOICE_NINJA_COLLECTION_VAT)
      ->get($vat->id());
    if (!$vatData) {
      // @todo Search for exising label, since it causes an error
      $remoteVat = $this->getInvoiceNinjaClient()->tax_rates->create(
        $this->getVatData($vat)
      );
      $this->storeId(self::INVOICE_NINJA_COLLECTION_VAT,
        $vat->id(), $remoteVat['data']['id']);
    }
    else {
      $lastSync = $vatData['last_sync'];
      // @phpstan-ignore-next-line
      if ($vat->getChangedTime() > $lastSync) {
        $remoteVatId = $vatData['remote_id'];
        $this->getInvoiceNinjaClient()->tax_rates->update($remoteVatId,
          $this->getVatData($vat)
        );
        $this->storeId(self::INVOICE_NINJA_COLLECTION_VAT,
          $vat->id(), $remoteVatId);
      }
    }
  }

  /**
   * Gets the data for the API call for a VAT.
   *
   * @param \Drupal\invoice_ninja\SyncableVatInterface $vat
   *   The VAT.
   *
   * @return array
   *   The VAT data for the API.
   */
  private function getVatData(SyncableVatInterface $vat): array {
    return [
      'name' => $vat->getVatLabel(),
      'rate' => $vat->getRate(),
    ];
  }

  /**
   * Syncs the invoice with Invoice Ninja.
   *
   * @param \Drupal\invoice_ninja\SyncableInvoiceInterface $invoice
   *   The invoice.
   */
  public function syncInvoice(SyncableInvoiceInterface $invoice): void {
    $remoteClientId = $this->syncClient($invoice->getClient());
    $invoiceData = $this->keyValueFactory->get(self::INVOICE_NINJA_COLLECTION_INVOICE)
      ->get($invoice->id());
    if (!$invoiceData) {
      $remoteInvoice = $this->getInvoiceNinjaClient()->invoices->create(
        $this->getInvoiceData($remoteClientId, $invoice)
      );
      $this->storeId(self::INVOICE_NINJA_COLLECTION_INVOICE,
        $invoice->id(), $remoteInvoice['data']['id']);
    }
    else {
      $lastSync = $invoiceData['last_sync'];
      // @phpstan-ignore-next-line
      if ($invoice->getChangedTime() > $lastSync) {
        $remoteInvoiceId = $invoiceData['remote_id'];
        $this->getInvoiceNinjaClient()->clients->update($remoteInvoiceId,
          $this->getInvoiceData($remoteClientId, $invoice)
        );
        $this->storeId(self::INVOICE_NINJA_COLLECTION_INVOICE,
          $invoice->id(), $remoteInvoiceId);
      }
    }
  }

  /**
   * Gets the data for the API call for an invoice for a given client.
   *
   * @param string $clientId
   *   The client ID.
   * @param \Drupal\invoice_ninja\SyncableInvoiceInterface $invoice
   *   The invoice.
   *
   * @return array
   *   The invoice data for the API.
   */
  private function getInvoiceData(string $clientId, SyncableInvoiceInterface $invoice): array {
    $lineItems = [];
    $isVatFree = $invoice->getClient()->isVatFree();
    foreach ($invoice->getLineItems() as $lineItem) {
      $vat = $lineItem->getVat();
      if ($vat !== NULL) {
        $this->syncVat($vat);
      }
      if ($isVatFree) {
        $vat = NULL;
      }
      $lineItems[] = [
        'notes' => $lineItem->getName(),
        'quantity' => $lineItem->getQuantity(),
        'cost' => $lineItem->getPrice(),
        'tax_name1' => $vat !== NULL ? $vat->getVatLabel() : '',
        'tax_rate1' => $vat !== NULL ? $vat->getRate() : 0,
      ];
    }
    return [
      'client_id' => $clientId,
      'line_items' => $lineItems,
    ];
  }

  /**
   * Gets an Invoice Ninja client.
   *
   * @return \InvoiceNinja\Sdk\InvoiceNinja
   *   The Invoice Ninja client.
   */
  private function getInvoiceNinjaClient(): InvoiceNinja {
    if (!isset($this->invoiceNinja)) {
      $url = $this->config->get(Settings::INVOICE_NINJA_URL);
      $token = $this->config->get(Settings::INVOICE_NINJA_API_TOKEN);
      $this->invoiceNinja = new InvoiceNinja($token);
      $this->invoiceNinja->setUrl($url);
    }
    return $this->invoiceNinja;
  }

  /**
   * Gets the mapping of countries.
   *
   * @return array
   *   The mapping of countries.
   */
  private function getCountryMapping(): array {
    if (empty($this->countryMapper)) {
      foreach ($this->getInvoiceNinjaClient()->statics->countries() as $country) {
        $this->countryMapper[$country['iso_3166_2']] = $country['id'];
      }
    }
    return $this->countryMapper;
  }

  /**
   * Gets the mapping of currencies.
   *
   * @return array
   *   The mapping of currencies.
   */
  private function getCurrencyMapping(): array {
    if (empty($this->currencyMapper)) {
      foreach ($this->getInvoiceNinjaClient()->statics->currencies() as $currency) {
        $this->currencyMapper[$currency['code']] = $currency['id'];
      }
    }
    return $this->currencyMapper;
  }

}
